@@ -51,7 +51,9 @@ typedef enum {
 
 static void gepub_doc_fill_resources (GepubDoc *doc);
 static void gepub_doc_fill_spine (GepubDoc *doc);
+static void gepub_doc_fill_toc (GepubDoc *doc, gchar *toc_id);
 static void gepub_doc_initable_iface_init (GInitableIface *iface);
+static gint navpoint_compare (GepubNavPoint *a, GepubNavPoint *b);
 
 struct _GepubDoc {
     GObject parent;
@@ -64,6 +66,7 @@ struct _GepubDoc {
 
     GList *spine;
     GList *chapter;
+    GList *toc;
 };
 
 struct _GepubDocClass {
@@ -103,6 +106,8 @@ gepub_doc_finalize (GObject *object)
     if (doc->spine) {
         g_list_foreach (doc->spine, (GFunc)g_free, NULL);
         g_clear_pointer (&doc->spine, g_list_free);
+        g_list_foreach (doc->toc, (GFunc)g_free, NULL);
+        g_clear_pointer (&doc->toc, g_list_free);
     }
 
     G_OBJECT_CLASS (gepub_doc_parent_class)->finalize (object);
@@ -198,6 +203,7 @@ gepub_doc_initable_init (GInitable     *initable,
     GepubDoc *doc = GEPUB_DOC (initable);
     gchar *file;
     gint i = 0, len;
+    g_autofree gchar *unescaped = NULL;
 
     g_assert (doc->path != NULL);
 
@@ -210,7 +216,8 @@ gepub_doc_initable_init (GInitable     *initable,
         }
         return FALSE;
     }
-    doc->content = gepub_archive_read_entry (doc->archive, file);
+    unescaped = g_uri_unescape_string (file, NULL);
+    doc->content = gepub_archive_read_entry (doc->archive, unescaped);
     if (!doc->content) {
         if (error != NULL) {
             g_set_error (error, gepub_error_quark (), GEPUB_ERROR_INVALID,
@@ -309,12 +316,19 @@ gepub_doc_fill_spine (GepubDoc *doc)
     const char *data;
     gsize size;
     GList *spine = NULL;
+    gchar *toc = NULL;
 
     data = g_bytes_get_data (doc->content, &size);
     xdoc = xmlRecoverMemory (data, size);
     root_element = xmlDocGetRootElement (xdoc);
     snode = gepub_utils_get_element_by_tag (root_element, "spine");
 
+    toc = gepub_utils_get_prop (snode, "toc");
+    if (toc) {
+        gepub_doc_fill_toc (doc, toc);
+        g_free (toc);
+    }
+
     item = snode->children;
     while (item) {
         if (item->type != XML_ELEMENT_NODE ) {
@@ -334,6 +348,107 @@ gepub_doc_fill_spine (GepubDoc *doc)
     xmlFreeDoc (xdoc);
 }
 
+static gint
+navpoint_compare (GepubNavPoint *a, GepubNavPoint *b)
+{
+    return a->playorder - b->playorder;
+}
+
+static void
+gepub_doc_fill_toc (GepubDoc *doc, gchar *toc_id)
+{
+    xmlDoc *xdoc = NULL;
+    xmlNode *root_element = NULL;
+    xmlNode *mapnode = NULL;
+    xmlNode *item = NULL;
+    const char *data;
+    gsize size;
+    GList *toc = NULL;
+    GBytes *toc_data = NULL;
+
+    doc->toc = toc;
+
+    toc_data = gepub_doc_get_resource_by_id (doc, toc_id);
+    if (!toc_data) {
+        return;
+    }
+
+    data = g_bytes_get_data (toc_data, &size);
+    xdoc = xmlRecoverMemory (data, size);
+    root_element = xmlDocGetRootElement (xdoc);
+    mapnode = gepub_utils_get_element_by_tag (root_element, "navMap");
+
+    // TODO: get docTitle
+    // TODO: parse metadata (dtb:totalPageCount, dtb:depth, dtb:maxPageNumber)
+
+    if (mapnode) {
+        item = mapnode->children;
+    }
+
+    while (item) {
+        GepubNavPoint *navpoint = NULL;
+        gchar *order;
+        xmlNode *navchilds = NULL;
+
+        if (item->type != XML_ELEMENT_NODE ||
+            g_strcmp0 ((const gchar *)item->name, "navPoint")) {
+            item = item->next;
+            continue;
+        }
+
+        navpoint = g_malloc0 (sizeof (GepubNavPoint));
+
+        order = gepub_utils_get_prop (item, "playOrder");
+        if (order) {
+            g_ascii_string_to_unsigned (order, 10, 0, INT_MAX,
+                                        &navpoint->playorder, NULL);
+            g_free (order);
+        }
+
+        // parsing navPoint->navLabel->text and navPoint->content
+        navchilds = item->children;
+        while (navchilds) {
+            if (item->type != XML_ELEMENT_NODE) {
+                navchilds = navchilds->next;
+                continue;
+            }
+
+            if (!g_strcmp0 ((const gchar *)navchilds->name, "content")) {
+                gchar **split;
+                gchar *tmpuri;
+                tmpuri = gepub_utils_get_prop (navchilds, "src");
+                // removing # params. Maybe we should store the # params in the
+                // navpoint to use in the future if the doc references to a position
+                // inside the chapter
+                split = g_strsplit (tmpuri, "#", -1);
+
+                // adding the base path
+                navpoint->content = g_strdup_printf ("%s%s", doc->content_base, split[0]);
+
+                g_strfreev (split);
+                g_free (tmpuri);
+            }
+
+            if (!g_strcmp0 ((const gchar *)navchilds->name, "navLabel")) {
+                xmlNode *text = gepub_utils_get_element_by_tag (navchilds, "text");
+                if (text->children && text->children->type == XML_TEXT_NODE) {
+                  navpoint->label = g_strdup ((gchar *)text->children->content);
+                }
+            }
+
+            navchilds = navchilds->next;
+        }
+
+        toc = g_list_prepend (toc, navpoint);
+        item = item->next;
+    }
+
+    doc->toc = g_list_sort (toc, (GCompareFunc) navpoint_compare);
+
+    xmlFreeDoc (xdoc);
+    g_bytes_unref (toc_data);
+}
+
 /**
  * gepub_doc_get_content:
  * @doc: a #GepubDoc
@@ -410,6 +525,7 @@ GBytes *
 gepub_doc_get_resource_by_id (GepubDoc *doc, const gchar *id)
 {
     GepubResource *gres;
+    g_autofree gchar *unescaped = NULL;
 
     g_return_val_if_fail (GEPUB_IS_DOC (doc), NULL);
     g_return_val_if_fail (id != NULL, NULL);
@@ -420,7 +536,8 @@ gepub_doc_get_resource_by_id (GepubDoc *doc, const gchar *id)
         return NULL;
     }
 
-    return gepub_archive_read_entry (doc->archive, gres->uri);
+    unescaped = g_uri_unescape_string (gres->uri, NULL);
+    return gepub_archive_read_entry (doc->archive, unescaped);
 }
 
 /**
@@ -433,10 +550,16 @@ gepub_doc_get_resource_by_id (GepubDoc *doc, const gchar *id)
 GBytes *
 gepub_doc_get_resource (GepubDoc *doc, const gchar *path)
 {
+    g_autofree gchar *unescaped = NULL;
+
     g_return_val_if_fail (GEPUB_IS_DOC (doc), NULL);
     g_return_val_if_fail (path != NULL, NULL);
 
-    return gepub_archive_read_entry (doc->archive, path);
+    // we need to decode the path because we can get URL encoded paths
+    // like "some%20text.jpg"
+    unescaped = g_uri_unescape_string (path, NULL);
+
+    return gepub_archive_read_entry (doc->archive, unescaped);
 }
 
 /**
@@ -475,15 +598,23 @@ gepub_doc_get_resource_mime (GepubDoc *doc, const gchar *path)
 {
     GepubResource *gres;
     GList *keys;
+    const gchar *_path;
 
     g_return_val_if_fail (GEPUB_IS_DOC (doc), NULL);
     g_return_val_if_fail (path != NULL, NULL);
 
+    if (path[0] == '/') {
+        _path = path + 1;
+    }
+    else {
+        _path = path;
+    }
+
     keys = g_hash_table_get_keys (doc->resources);
 
     while (keys) {
         gres = ((GepubResource*)g_hash_table_lookup (doc->resources, keys->data));
-        if (!strcmp (gres->uri, path))
+        if (!strcmp (gres->uri, _path))
             break;
         keys = keys->next;
     }
@@ -547,6 +678,7 @@ gepub_doc_get_current_with_epub_uris (GepubDoc *doc)
 
     replaced = gepub_utils_replace_resources (content, base);
 
+    g_free (base);
     g_free (path);
     g_bytes_unref (content);
 
@@ -803,3 +935,99 @@ gepub_doc_get_current_id (GepubDoc *doc)
 
     return doc->chapter->data;
 }
+
+/**
+ * gepub_doc_get_toc:
+ * @doc: a #GepubDoc
+ *
+
+ * Returns: (element-type Gepub.NavPoint) (transfer none): the navigation list in order
+ */
+GList *
+gepub_doc_get_toc (GepubDoc *doc)
+{
+    g_return_val_if_fail (GEPUB_IS_DOC (doc), NULL);
+    return doc->toc;
+}
+
+/**
+ * gepub_doc_resource_uri_to_chapter:
+ * @doc: a #GepubDoc
+ * @uri: The resource path
+ *
+ * This method tries to find the resource by path in the doc spine and
+ * will return the index in that list. If the resourse isn't there this method
+ * will return -1.
+
+ * Returns: the chapter index to use with gepub_doc_set_chapter or -1 if the
+ * resource isn't found
+ */
+gint
+gepub_doc_resource_uri_to_chapter (GepubDoc *doc,
+                                   const gchar *uri)
+{
+    GHashTableIter iter;
+    gchar *key;
+    GepubResource *res;
+    gchar *id = NULL;
+    const gchar *_uri;
+
+    if (uri[0] == '/') {
+        _uri = uri + 1;
+    }
+    else {
+        _uri = uri;
+    }
+
+    g_return_val_if_fail (GEPUB_IS_DOC (doc), -1);
+    g_return_val_if_fail (doc->spine != NULL, -1);
+
+    g_hash_table_iter_init (&iter, doc->resources);
+    while (g_hash_table_iter_next (&iter, (gpointer *)&key, (gpointer *)&res)) {
+        if (!g_strcmp0 (res->uri, _uri)) {
+            id = key;
+            break;
+        }
+    }
+
+    if (!id) {
+        return -1;
+    }
+
+    return gepub_doc_resource_id_to_chapter (doc, id);
+}
+
+/**
+ * gepub_doc_resource_id_to_chapter:
+ * @doc: a #GepubDoc
+ * @id: The resource id
+ *
+ * This method tries to find the resource by id in the doc spine and
+ * will return the index in that list. If the resourse isn't there this method
+ * will return -1.
+
+ * Returns: the chapter index to use with gepub_doc_set_chapter or -1 if the
+ * resource isn't found
+ */
+gint
+gepub_doc_resource_id_to_chapter (GepubDoc *doc,
+                                  const gchar *id)
+{
+    GList *spine;
+    gint chapter = 0;
+
+    g_return_val_if_fail (GEPUB_IS_DOC (doc), -1);
+    g_return_val_if_fail (doc->spine != NULL, -1);
+
+    spine = g_list_first (doc->spine);
+    while (spine && spine->data) {
+        if (!g_strcmp0 (spine->data, id)) {
+            return chapter;
+        }
+        chapter++;
+        spine = spine->next;
+    }
+
+    return -1;
+}
+
